#define noid_root 4007
#define noid_blok 5113
#define noid_free 5433
#define noid_lost 1057

typedef unsigned long int klop_id;

typedef struct NOiD_klop
{
	unsigned long int id;		//permanent id , represent the real place in the array
	unsigned long int nfo;		//check sum id , defined higher
	unsigned long int head;		//continuous file
	unsigned long int tail;		//continuous file
	unsigned long int last;
	unsigned long int next;
	unsigned long int back;
	unsigned long int fore;
} NOiD_klop;

//===============================================================
//===============================================================
//                     FORMAT
//===============================================================
//===============================================================
NOiD_klop * NOiD_build (int noid_qty, int noid_cluster, void * noid_ptr)
{
	if ((noid_qty == 0)||(noid_ptr == 0)||(noid_cluster <= sizeof(NOiD_klop)))
	{	return 0;	}

	int noid_tsize = noid_qty * sizeof(NOiD_klop);
	int noid_tend = (noid_tsize + (sizeof(NOiD_klop)-1)) / noid_cluster;
	int noid_start = noid_tend + 1;
	int noid_end = noid_qty;
	int noid_go = 0;
	NOiD_klop * noid_table = (NOiD_klop*) noid_ptr;
	NOiD_klop * noid_libre = (NOiD_klop*)((noid_ptr) + (noid_start * sizeof(NOiD_klop)));

	while (noid_go <= noid_qty)
	{
		NOiD_klop * noid_new = (NOiD_klop*)(noid_ptr + (sizeof(NOiD_klop) * noid_go));
		noid_new->id = noid_go;
		if (noid_go >= noid_start)
		{
			noid_new->nfo = noid_free;
			if (noid_go == noid_start)
			{	noid_new->back = noid_end;	}
			else
			{	noid_new->back = noid_go - 1;	}
			if (noid_go == noid_end)
			{	noid_new->fore = noid_start;	}
			else
			{	noid_new->fore = noid_go + 1;	}
			noid_new->last = noid_new->id;
			noid_new->next = noid_new->id;
		}
		else
		{
			noid_new->nfo = noid_root;
			if (noid_go == 0)
			{	noid_new->back = noid_tend;	}
			else
			{	noid_new->back = noid_go - 1;	}
			if (noid_go == noid_tend)
			{	noid_new->fore = 0;		}
			else
			{	noid_new->fore = noid_go + 1;	}
			noid_new->last = noid_new->id;
			noid_new->next = noid_new->id;
		}
		noid_go += 1;
	}
	noid_table->last = noid_start;
	noid_table->next = noid_start;
	noid_libre->last = 0;
	noid_libre->next = 0;

	return noid_table;
}
//===============================================================
//===============================================================
//                INTERNAL  FUNCTION
//                 FOR LIGHTER CODE
//===============================================================
//===============================================================
NOiD_klop * NOiD_getklop (NOiD_klop * noid_table, int noid_numb)
/* internal function */
{
	return (NOiD_klop*)(((void*)noid_table) + (noid_numb * sizeof(NOiD_klop)));
}

//===============================================================
//===============================================================
//                ORTHOGONAL PERSISTENCE
//                  @ CLUSTER LEVEL
//===============================================================
//===============================================================

int NOiD_pop (NOiD_klop * noid_table, int noid_src)
/* will pop blok to free while keeping undo redo , then 1st redo will take place
   noid_table = pointer to the table in memory
   noid_src = cluster permanent id of the blok */
{
	if ((noid_table == 0)||(noid_table->id != 0)||(noid_src == 0))
	{ 	return 999999;	}

	NOiD_klop * noid_here = NOiD_getklop (noid_table, noid_src);
	NOiD_klop * noid_next = NOiD_getklop (noid_table, noid_here->next);
	NOiD_klop * noid_last = NOiD_getklop (noid_table, noid_here->last);
	NOiD_klop * noid_back = NOiD_getklop (noid_table, noid_here->back);
	NOiD_klop * noid_fore = NOiD_getklop (noid_table, noid_here->fore);

	if ((noid_here->back == noid_here->id)||(noid_here->fore == noid_here->id))
	{
		noid_next->last = noid_here->last;
		noid_last->next = noid_here->next;
	}
	else
	{
		noid_next->last = noid_fore->id;
		noid_last->next = noid_fore->id;
		noid_back->fore = noid_here->fore;
		noid_fore->back = noid_here->back;
		noid_here->back = noid_here->id;
		noid_here->fore = noid_here->id;
		noid_fore->last = noid_last->id;
		noid_fore->next = noid_next->id;
	}
	noid_here->last = noid_here->id;
	noid_here->next = noid_here->id;

	noid_here->nfo = noid_lost;

	return noid_here->id;
}

int NOiD_push (NOiD_klop * noid_table, int noid_src, int noid_dest)
/* will push ure blok under undo list and then add new free cluster to blok
   noid_table = pointer to the table in memory
   noid_src = cluster permanent id of the blok */
{
	if ((noid_table == 0)||(noid_table->id != 0)||(noid_src == 0))
	{ 	return;	}

	NOiD_klop * noid_new = NOiD_getklop (noid_table, noid_src);
	NOiD_klop * noid_next, * noid_last;

	if (noid_dest != 0)
	{
		NOiD_klop * noid_here = NOiD_getklop (noid_table, noid_dest);
		NOiD_klop * noid_fore = NOiD_getklop (noid_table, noid_here->fore);
		noid_new->back = noid_here->id;
		noid_new->fore = noid_fore->id;
		noid_here->fore = noid_new->id;
		noid_fore->back = noid_new->id;
		noid_new->nfo = noid_here->nfo;
		noid_next = NOiD_getklop (noid_table, noid_here->next);
		noid_last = NOiD_getklop (noid_table, noid_here->last);
		noid_here->last = noid_here->id;
		noid_here->next = noid_here->id;
	}
	else
	{
		noid_next = NOiD_getklop (noid_table, noid_table->next);
		noid_last = NOiD_getklop (noid_table, noid_table->id);
		noid_new->nfo = 0;
	}

	noid_next->last = noid_new->id;
	noid_last->next = noid_new->id;
	noid_new->last = noid_last->id;
	noid_new->next = noid_next->id;

	return noid_src;
}

int NOiD_last (NOiD_klop * noid_table, int noid_src)
{
	if ((noid_table == 0)||(noid_table->id != 0))
	{ 	return 999999;	}

	NOiD_klop * noid_here = NOiD_getklop (noid_table, noid_src);
	return noid_here->last;
}

int NOiD_next (NOiD_klop * noid_table, int noid_src)
{
	if ((noid_table == 0)||(noid_table->id != 0))
	{ 	return 999999;	}

	NOiD_klop * noid_here = NOiD_getklop (noid_table, noid_src);
	return noid_here->next;
}

int NOiD_undo (NOiD_klop * noid_table, int noid_src)
/* will rotate blok list to give next undo first access
   noid_table = pointer to the table in memory
   noid_src = cluster permanent id of the blok */
{
	if ((noid_table == 0)||(noid_table->id != 0)||(noid_src == 0)||(noid_src == 1))
	{ 	return 999999;	}

	NOiD_klop * noid_here = NOiD_getklop (noid_table, noid_src);
	NOiD_klop * noid_undo = NOiD_getklop (noid_table, noid_here->back);
	NOiD_klop * noid_last = NOiD_getklop (noid_table, noid_here->last);
	NOiD_klop * noid_next = NOiD_getklop (noid_table, noid_here->next);

	noid_here->last = noid_here->id;
	noid_here->next = noid_here->id;
	noid_last->next = noid_undo->id;
	noid_next->last = noid_undo->id;
	noid_undo->last = noid_last->id;
	noid_undo->next = noid_next->id;

	return noid_undo->id;
}

int NOiD_redo (NOiD_klop * noid_table, int noid_src)
/* will rotate blok list to give next redo first access
   noid_table = pointer to the table in memory
   noid_src = cluster permanent id of the blok */
{
	if ((noid_table == 0)||(noid_table->id != 0)||(noid_src == 0)||(noid_src == 1))
	{ 	return 999999;	}

	NOiD_klop * noid_here = NOiD_getklop (noid_table, noid_src);
	NOiD_klop * noid_redo = NOiD_getklop (noid_table, noid_here->fore);
	NOiD_klop * noid_last = NOiD_getklop (noid_table, noid_here->last);
	NOiD_klop * noid_next = NOiD_getklop (noid_table, noid_here->next);

	noid_here->last = noid_here->id;
	noid_here->next = noid_here->id;
	noid_last->next = noid_redo->id;
	noid_next->last = noid_redo->id;
	noid_redo->last = noid_last->id;
	noid_redo->next = noid_next->id;

	return noid_redo->id;
}

int NOiD_add(NOiD_klop * noid_table,int noid_head, int noid_new)
/* add given new kluster to the end of the file */
/* then return begining of the file */
{

	return 0;
}

int NOiD_sub(NOiD_klop * noid_table,int noid_head)
/* eleminate last kluster of the file */
/* then return errant kluster */
{

	return 0;
}

//===============================================================
//===============================================================
//                    4 ramdrive
//===============================================================
//===============================================================
void * NOiD_read (NOiD_klop * noid_table, int noid_size, int noid_numb, void * noid_dest)
{
	int noid_down = noid_size;
	while (noid_down != 0)
	{
		*(char*)(noid_dest + noid_down) = *(char*)(((void*)noid_table) + (noid_numb * noid_size) + noid_down);
		noid_down -= 1;
	}
	return noid_dest;
}

void * NOiD_write (NOiD_klop * noid_table, int noid_size, int noid_numb, void * noid_src)
{
	int noid_down = noid_size;
	while (noid_down != 0)
	{
		*(char*)(((void*)noid_table) + (noid_numb * noid_size) + noid_down) = *(char*)(noid_src + noid_down);
		noid_down -= 1;
	}
	return noid_src;
}
